from __future__ import annotations
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Dict,
    List,
)

from mongoengine import context_managers
from stereotech_cloud_commons.components.mongo_database import Role, RoleEntity

if TYPE_CHECKING:
    from stereotech_cloud_commons.confighelper import ConfigHelper
    from stereotech_cloud_commons.websockets import WebRequest


class Roles:
    def __init__(self, config: ConfigHelper) -> None:
        """Service for managing user roles."""
        self.server = config.get_server()
        self.entity_name = config.get('entity_name', None)
        self.collection_name = f'{self.entity_name}_roles' if self.entity_name is not None else 'roles'
        self.context = context_managers.switch_collection(
            Role, self.collection_name)

        logging.info("Roles configuration loaded.")

        self.server.register_endpoint("/roles/role", ["POST", "DELETE", "GET"],
                                      self.handle_manage_role)
        self.server.register_endpoint("/roles/list", ["GET"],
                                      self.handle_all_roles)
        self.server.register_endpoint("/roles/scopes", ["GET"],
                                      self.handle_all_scopes)

    async def handle_all_roles(self, web_request: WebRequest) -> Dict[str, Any]:
        args = web_request.get_args()
        query = Role
        count = 0
        items = []
        if self.entity_name is not None:
            entity_id = web_request.get_str(f'{self.entity_name}_id')
            with self.context as Role_ctx:
                query = Role_ctx.objects(entity__id=entity_id)  # type: ignore
                count, items = query.get_items(**args)  # type: ignore
        else:
            count, items = query.get_items(**args)  # type: ignore
        return {"count": count, "items": items}

    async def handle_manage_role(self, web_request: WebRequest) -> Role | Dict[str, Any]:
        action = web_request.get_action()
        if self.context is not None:
            with self.context as Role:
                if action == "POST":
                    data = web_request.get_args()

                    # Update role
                    if "id" in data:
                        logging.info("Update role | id: %s.", data["id"])
                        data.pop("entity", None)
                        role = self.edit_role(data)
                        return role

                    # Create role
                    else:
                        name = web_request.get_str('name')
                        scopes = web_request.get('scopes')
                        logging.info("Creating role | name: %s, scopes: %s.", name,
                                     scopes)
                        entity_id = ''
                        if self.entity_name is not None:
                            entity_id = web_request.get(
                                f'{self.entity_name}_id')
                        role = self.create_role(name, scopes, entity_id)
                        return role

                elif action == "DELETE":
                    role_id: str = web_request.get_str("id")
                    logging.info("Delete role | id: %s.", role_id)
                    self.delete_role(role_id)
                    return {"deleted": role_id}

                elif action == "GET":
                    role_id: str = web_request.get_str("id")
                    logging.info("Get role | id: %s.", role_id)
                    return self.get_role_or_404(role_id)

        raise self.server.error('Method Not Allowed', 405)

    async def handle_all_scopes(self, web_request: WebRequest) -> Dict[str, Any]:
        """Get a list of all scopes."""
        args = web_request.get_args()
        with self.context as Role:
            count, items = Role.get_scopes(**args)  # type: ignore
        return {"count": count, "items": items}

    def get_role_or_404(self, role_id: str) -> Role:
        with self.context as Role:
            try:
                role_db = Role.get_item(role_id)  # type: ignore
            except:
                raise self.server.error(f"Role {role_id} not found", 404)
            if role_db is None:
                raise self.server.error(f"Role {role_id} not found", 404)
            return role_db

    def query_role(self, **query):
        with self.context as Role:
            return Role.objects(**query)

    def create_role(self, name: str, scopes: List[str], entity_id: str = ''):
        with self.context as Role:
            role = Role(name=name, scopes=scopes)
            if self.entity_name is not None:
                role.entity = RoleEntity(name=self.entity_name, id=entity_id)
            role.save()
            return role

    def delete_role(self, role_id):
        with self.context as Role:
            try:
                role_db = Role.get_item(role_id)  # type: ignore
            except:
                raise self.server.error(f"Role {role_id} not found", 404)
            if role_db is None:
                raise self.server.error(f"Role {role_id} not found", 404)
            role_db.delete()

    def edit_role(self, data):
        with self.context as Role:
            try:
                role_db = Role.get_item(data["id"])  # type: ignore
            except:
                raise self.server.error(f"Role {data["id"]} not found", 404)
            if role_db is None:
                raise self.server.error(f"Role {data["id"]} not found", 404)
            del data["id"]
            role_db.modify(**data)
            role_db.save()
            return role_db


def load_component(config: ConfigHelper) -> Roles:
    return Roles(config)



import datetime
import math
import random
import time
from typing import Optional
from mongoengine import *
from mongoengine import signals
from operator import itemgetter


def handler(event):
    """Signal decorator to allow use of callback functions as class decorators."""
    def decorator(fn):
        def apply(cls):
            event.connect(fn, sender=cls)
            return cls

        fn.apply = apply
        return fn

    return decorator


@handler(signals.pre_save)
def update_date_tracking(sender, document):
    document.update_date = datetime.datetime.utcnow()


class ListQuerySet(QuerySet):
    def get_items(self, search: Optional[str] = None, sort_by: str = 'id', sort_desc: bool = False, skip: int = 0, limit: int = 10, **kwargs):
        query = self.order_by(
            f"{'-' if sort_desc else '+'}{sort_by}")[skip:limit]
        if search is not None:
            query = query.search_text(search)
        return query.count(), [i.to_mongo() for i in query]

    def get_item(self, id: str):
        return self.get(id=id)


class ItemFrequenciesQuerySet(ListQuerySet):
    def get_field_frequencies(self, field: str,  search: Optional[str] = None, sort_desc: bool = True, skip: int = 0, limit: int = 100, **kwargs):
        frqus = self.item_frequencies(field)
        if search is not None and len(search) > 0:
            frqus = {k: v for k, v in frqus.items() if search in str(k)}
        frqus_list = sorted(frqus.items(), key=itemgetter(1),
                            reverse=sort_desc)[skip:limit]
        return len(frqus_list), [{'id': kv[0], 'count': kv[1]} for kv in frqus_list]


class ListableDocument(Document):
    meta = {'allow_inheritance': True, 'queryset_class': ListQuerySet, 'abstract': True}
    creation_date = DateTimeField(default=datetime.datetime.utcnow())

    @queryset_manager
    def get_items(doc_cls, queryset: QuerySet, search: Optional[str] = None, sort_by: str = 'id', sort_desc: bool = False, skip: int = 0, limit: int = 100, **kwargs):
        return queryset.get_items(search, sort_by, sort_desc, skip, limit, **kwargs)

class RoleEntity(EmbeddedDocument):
    name = StringField(required=True)
    id = StringField(required=True)

class Role(ListableDocument):
    meta = {'allow_inheritance': True, 'queryset_class': ItemFrequenciesQuerySet, 'collection': 'roles'}

    entity = EmbeddedDocumentField(RoleEntity)
    name = StringField(required=True)
    scopes = ListField(StringField())

    @queryset_manager
    def get_scopes(doc_cls, queryset: QuerySet, search: Optional[str] = None, sort_desc: bool = True, skip: int = 0, limit: int = 100, **kwargs):
        return queryset.get_field_frequencies(
            'scopes', search, sort_desc, skip, limit, **kwargs)


class UserVerification(EmbeddedDocument):
    verified = BooleanField(default=False)
    verification_code = IntField(
        1000, 9999, default=random.randrange(1000, 9999))


class JwtData(EmbeddedDocument):
    jwt_id = StringField(required=True)
    jwt_secret = StringField(required=True)



class User(ListableDocument):
    meta = {'collection': 'users'}

    email = EmailField(required=True, unique=True)
    password = StringField(required=True)
    username = StringField(min_length=1, max_length=50)
    avatar = URLField()
    role = ReferenceField(Role)
    verification = EmbeddedDocumentField(UserVerification)
    jwt_data = EmbeddedDocumentField(JwtData)
    salt = StringField(required=True)

    @queryset_manager
    def get_user(doc_cls, queryset: ListQuerySet, id: str):
        return queryset.get_item(id)

    @queryset_manager
    def get_users(doc_cls, queryset: ListQuerySet, **kwargs):
        return queryset.get_items(**kwargs)[1].exclude('verification', 'jwt_data')



import json
import logging

from ..confighelper import ConfigHelper
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.httputil import HTTPHeaders
from typing import Any, Dict


class Storage:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.storage_server = config.get('storage_server', None)
        self.client = AsyncHTTPClient()
        if self.storage_server is None:
            raise config.error("storage_server is required")

    async def create_dir(self, path: str, headers: HTTPHeaders) -> Any:
        token: str = headers.get("Authorization").split(" ")[1]
        server_path: str = "/server/files/directory"
        request = HTTPRequest(
            f"{self.storage_server}{server_path}", 
            "POST", 
            body=json.dumps({
                "path": path,
            }),
            headers={
                "Authorization": "Bearer " + token,
                "Content-Type": "application/json",
            },
        )
        try: 
            result = await self.client.fetch(request)
            return json.loads(result.body)
        except Exception as e:
            logging.info("Can't create a new folder: %s", e)
            return None
        
    async def get_metadata(self, filename: str, headers: HTTPHeaders) -> Dict[str, Any]:
        token: str = headers.get("Authorization").split(" ")[1]
        server_path: str = "/server/files/metadata"
        request = HTTPRequest(
            f"{self.storage_server}{server_path}",
            method="GET",
            body=json.dumps({
                "filename": filename,
            }),
            headers={
                "Authorization": "Bearer " + token,
                "Content-Type": "application/json",
            },
            allow_nonstandard_methods=True,
        )
        try: 
            result = await self.client.fetch(request)
            return json.loads(result.body)["result"]
        except Exception as e:
            logging.info("Can't find metadata: %s", e)
            return None


def load_component(config: ConfigHelper) -> Storage:
    return Storage(config)
